"""
Main command line interface of the phlorest package.

Like programs such as git, this cli splits its functionality into sub-commands
(see e.g. https://docs.python.org/2/library/argparse.html#sub-commands).
The rationale behind this is that while a lot of different tasks may be
triggered using this cli, most of them require common configuration.
"""
import sys
import argparse
import contextlib

import phlorest

from clldutils.clilib import register_subcommands
from clldutils.clilib import get_parser_and_subparsers
from clldutils.clilib import ParserError
from clldutils.loglib import Logging

from cldfcatalog import Config
from cldfbench.catalogs import BUILTIN_CATALOGS

import termcolor


def main(args=None, catch_all=False, parsed_args=None, log=None):  # pragma: no cover
    parser, subparsers = get_parser_and_subparsers(phlorest.__name__)

    # We add a "hidden" option to turn-off config file reading in tests:
    parser.add_argument('--no-config', default=False, action='store_true', help=argparse.SUPPRESS)

    # Discover available commands:
    # Commands are identified by (<entry point name>).<module name>
    register_subcommands(subparsers, phlorest.commands)

    args = parsed_args or parser.parse_args(args=args)
    if not hasattr(args, "main"):
        parser.print_help()
        return 1

    with contextlib.ExitStack() as stack:
        if not log:  # pragma: no cover
            stack.enter_context(Logging(args.log, level=args.log_level))
        else:
            args.log = log

        # args.no_catalogs is set by the `config` command, because this command specifies
        # catalog options **optionally**, and prompts for user input only in its `run` function.
        if not getattr(args, "no_catalogs", False):
            cfg = Config.from_file()
            for cls in BUILTIN_CATALOGS:
                # Now we loop over known catalogs, see whether they are used by the command,
                # and if so, "enter" the catalog.
                name, from_cfg = cls.cli_name(), False
                if hasattr(args, name):
                    # If no path was passed on the command line, we look up the config:
                    path = getattr(args, name)
                    if (not path) and (not args.no_config):
                        try:
                            path = cfg.get_clone(name)
                            from_cfg = True
                        except KeyError as e:  # pragma: no cover
                            print(termcolor.colored(str(e) + '\n', 'red'))
                            return main([args._command, '-h'])
                    try:
                        setattr(
                            args,
                            name,
                            stack.enter_context(
                                cls(path, getattr(args, name + '_version', None))),
                        )
                    except ValueError as e:
                        print(termcolor.colored(
                            '\nError initializing catalog {0}'.format(name), 'red'))
                        if from_cfg:
                            print(
                                termcolor.colored('from config {0}'.format(cfg.fname()), 'red'))
                        print(termcolor.colored(str(e) + '\n', 'red'))
                        return main([args._command, '-h'])

        try:
            return args.main(args) or 0
        except KeyboardInterrupt:  # pragma: no cover
            return 0
        except ParserError as e:
            print(termcolor.colored('ERROR: {}\n'.format(e), 'red', attrs={'bold'}))
            return main([args._command, '-h'])
        except Exception as e:
            if catch_all:  # pragma: no cover
                print(termcolor.colored('ERROR: {}\n'.format(e), 'red', attrs={'bold'}))
                return 1
            raise


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main() or 0)



"""
Checks datasets for compliance
"""
from termcolor import colored
from cldfbench.cli_util import add_dataset_spec, get_dataset

from phlorest.check import run_checks


def register(parser):  # pragma: no cover
    add_dataset_spec(parser)


def run(args, d=None):
    if d is None:  # pragma: no cover
        try:
            d = get_dataset(args)
        except Exception as e:
            args.log.error("Unable to load %s - %s" % (args.dataset, e))
            raise

    msg, color = ('PASS', 'green') if run_checks(d, args.log) else ('FAIL', 'red')
    print('{} {}'.format(colored(msg, color, attrs=['bold']), d.id))



from phlorest import Metadata
from phlorest.scaffold import PhlorestTemplate


def test_template(tmp_path):
    PhlorestTemplate().render(tmp_path, Metadata(id='x', name='n', year=2021, author='a'))
    assert tmp_path.joinpath('x', 'etc', 'taxa.csv').exists()



"""

"""
import collections

from csvw.dsv import reader, UnicodeWriter
from clldutils.clilib import PathType
from cldfbench.cli_util import add_catalog_spec, add_dataset_spec, get_dataset


def register(parser):  # pragma: no cover
    add_catalog_spec(parser, 'concepticon')
    add_dataset_spec(parser)
    parser.add_argument('mapping', type=PathType(type='file'))


def run(args):  # pragma: no cover
    ds = get_dataset(args)

    conceptsets = {cs.gloss: cs for cs in args.concepticon.api.conceptsets.values()}

    chars = collections.OrderedDict([(c['Site'], c) for c in ds.characters])
    for mapping in reader(args.mapping, dicts=True):
        if mapping['Site'] not in chars:
            args.log.info('adding {}'.format(mapping))
            chars[mapping['Site']] = collections.OrderedDict(
                [('Site', mapping['Site']), ('Label', mapping['Label'])])
        else:
            if 'Label' in mapping and 'Label' in chars[mapping['Site']]:
                assert mapping['Label'] == chars[mapping['Site']]['Label']

        assert mapping['Word'] in conceptsets
        chars[mapping['Site']]['concepticonReference'] = conceptsets[mapping['Word']].id
        chars[mapping['Site']]['Concepticon_Gloss'] = mapping['Word']

    with UnicodeWriter(ds.etc_dir / 'characters.csv') as w:
        for i, char in enumerate(chars.values()):
            if i == 0:
                cols = list(char.keys())
                if 'concepticonReference' not in cols:
                    cols.append('concepticonReference')
                if 'Concepticon_Gloss' not in cols:
                    cols.append('Concepticon_Gloss')
                w.writerow(cols)
            w.writerow([char.get(col, '') for col in cols])



from phlorest.metadata import Metadata


def test_Metadata():
    assert Metadata(id='abc').title == 'Phlorest phylogeny abc'
    md = Metadata(
        url='http://dx.doi.org/a/b',
        name='the name',
        scaling='centuries',
        analysis='other',
        data='data.nex',
        author='the author',
        year='2021')
    assert md.url == 'https://doi.org/a/b'
    assert md.common_props()



import typing
import pathlib

import cldfbench
import newick
import tqdm
from pycldf.terms import TERMS
from commonnexus import Nexus
from commonnexus.tools.normalise import normalise
from commonnexus.tools.matrix import CharacterMatrix
from commonnexus.blocks.characters import Characters

from .beast import BeastFile
from .metadata import Metadata
from .nexuslib import NexusFile


class CLDFWriter(cldfbench.CLDFWriter):
    """
    A CLDF writer that knows how to add phylogentic data.
    """
    def __enter__(self):
        self._lids = set()
        self.summary = NexusFile(self.cldf_spec.dir / 'summary.trees')
        self.summary.__enter__()
        self.posterior = NexusFile(self.cldf_spec.dir / 'posterior.trees', zipped=True)
        self.posterior.__enter__()
        res = cldfbench.CLDFWriter.__enter__(self)
        self.add_schema()
        return res

    def __exit__(self, *args):
        if self.dataset:
            if self.dataset.metadata.cldf:
                self.cldf.add_provenance(
                    wasDerivedFrom={
                        "rdf:about": self.dataset.metadata.cldf,
                        "rdf:type": "prov:Entity",
                        'dc:description': 'The CLDF dataset from which the data underlying the '
                                          'analysis was derived',
                        'dc:format': 'https://cldf.clld.org',
                    }
                )
            if self.dataset.metadata.data and self.dataset.metadata.data.startswith('http'):
                self.cldf.add_provenance(
                    wasDerivedFrom={
                        "rdf:about": self.dataset.metadata.data,
                        "rdf:type": "prov:Entity",
                        'dc:description': 'The dataset from which the data underlying the '
                                          'analysis was derived',
                    }
                )

        self.summary.__exit__(*args)
        self.posterior.__exit__(*args)
        return cldfbench.CLDFWriter.__exit__(self, *args)

    def add_schema(self):
        t = self.cldf.add_component('LanguageTable')
        t.common_props['dc:description'] = \
            "The LanguageTable lists the taxa, i.e. the leafs of the phylogeny, mapped to " \
            "languoids."
        self.cldf.add_component('TreeTable')
        self.cldf.add_component('MediaTable')

    def add_columns(self, table, obj, log, exclude=None):
        """
        Wraps `pycldf.Dataset.add_columns`, adding some checking.
        """
        existing = [c.name for c in self.cldf[table].tableSchema.columns]
        exclude = exclude or []
        new = []
        for k in obj.keys():
            lname = 'concepticonReference' if k == 'Concepticon_ID' else k
            if k not in exclude:
                col = TERMS[lname].to_column() if lname in TERMS else k
                if getattr(col, 'name', k) in existing:
                    log.error('Duplicate column name {} for {}'.format(k, table))
                    continue
                new.append(col)

        self.cldf.add_columns(table, *new)

    def add_obj(self,
                table: str,
                d: dict,
                row: typing.Optional[dict] = None,
                rename: typing.Optional[typing.Dict[str, str]] = None):
        """
        Merge data from `row` into `d` and add the resulting `dict` to table `table`.
        """
        rename = rename or {}
        for k, v in (row or {}).items():
            k = rename.get(k, k)
            if k in TERMS:
                k = TERMS[k].to_column().name
            d[k] = v
        self.objects[table].append(d)

    def add_tree(self,
                 tree: typing.Union[str, newick.Node],
                 nex: NexusFile,
                 tid: str,
                 metadata: Metadata,
                 log,
                 type_,
                 source=None,
                 rooted: typing.Optional[bool] = None):
        nex.append(tree, tid, self._lids, metadata.scaling, log, rooted=rooted)
        if source is None:
            bibkeys = list(self.cldf.sources.keys())
            if len(bibkeys) == 1:
                source = bibkeys[0]

        # Add media file only if necessary!
        mids = [m['ID'] for m in self.objects['MediaTable']]
        if nex.path.stem not in mids:
            self.objects['MediaTable'].append(dict(
                ID=nex.path.stem,
                Media_Type='text/plain',
                Download_URL='file:///{}{}'.format(
                    nex.path.name, '' if nex.path.stem == 'summary' else '.zip'),
                Path_In_Zip=None if nex.path.stem == 'summary' else 'posterior.trees',
            ))

        self.objects['TreeTable'].append(dict(
            ID=tid,
            Name=tid,
            Media_ID=nex.path.stem,
            Tree_Is_Rooted=rooted,
            Tree_Type=type_,
            Description=metadata.analysis,
            Tree_Branch_Length_Unit=None if nex.scaling == 'none' else nex.scaling,
            Source=[source] if isinstance(source, str) else source,
        ))

    def add_summary(self,
                    tree: typing.Union[str, newick.Node],
                    metadata: Metadata,
                    log,
                    source=None,
                    rooted=None):
        """
        Add `tree` as summary tree to the dataset.
        """
        self.add_tree(
            tree, self.summary, 'summary', metadata, log, 'summary', source=source, rooted=rooted)
        log.info("added summary tree")

    def add_posterior(self,
                      trees: typing.List[typing.Union[str, newick.Node]],
                      metadata: Metadata,
                      log,
                      source=None,
                      verbose=False,
                      rooted=None):
        """
        Add `trees` as posterior sample of trees to the dataset.
        """
        i = 0
        for i, tree in (
                tqdm.tqdm(enumerate(trees, start=1), total=len(trees))
                if verbose else enumerate(trees, start=1)):
            self.add_tree(
                tree,
                self.posterior,
                'posterior-{}'.format(i),
                metadata,
                log,
                'sample',
                source=source,
                rooted=rooted)
        log.info("added posterior trees (n=%d)" % i)

    def add_data(self,
                 input: typing.Union[BeastFile, pathlib.Path, str, Nexus],
                 characters: typing.Iterable[typing.Dict[str, str]],
                 log,
                 binarise: bool = False):
        """
        Add character data from which the tree(s) in the dataset were computed.

        :param input: Character data can be read from BEAST files and NEXUS files.
        :param characters: Character metadata, per site.
        :param log:
        """
        if isinstance(input, BeastFile):
            nex = input.nexus()
        elif isinstance(input, pathlib.Path):
            nex = Nexus.from_file(input)
        elif isinstance(input, str):
            nex = Nexus(input)
        else:
            nex = input
        assert isinstance(nex, Nexus)
        charlabels, _ = nex.characters.get_charstatelabels()

        md = {int(row.pop('Site')): row for row in characters}
        t = self.cldf.add_component(
            'ParameterTable',
            {
                'name': 'Nexus_File',
                'dc:description':
                    'The data for this parameter is stored at 1-based index {ID} '
                    'of the sequences in the DATA block of the Nexus file specified here. '
                    '(See https://en.wikipedia.org/wiki/Nexus_file)',
                'propertyUrl': 'http://cldf.clld.org/v1.0/terms.rdf#mediaReference',
            },
        )
        t.common_props['dc:description'] = \
            "The ParameterTable lists characters (a.k.a. sites), i.e. the (often binary) " \
            "variables used as data basis to compute the phylogeny from."
        if md:
            self.add_columns(
                'ParameterTable', list(md.values())[0], log, exclude=['Label'])
        self.cldf['ParameterTable', 'ID'].common_props['dc:description'] = \
            "Sequence index of the site in the corresponding Nexus file."
        for site, label in charlabels.items():
            d = dict(ID=site, Name=label, Nexus_File='data')
            self.add_obj('ParameterTable', d, md.get(site, {}), rename=dict(Label='Name'))
        self.add_obj(
            'MediaTable',
            dict(ID='data', Media_Type='text/plain', Download_URL='file:///data.nex'))

        assert all(t in self._lids for t in nex.taxa), "Taxa in nexus not in taxa.csv: {}".format(
            [t for t in nex.taxa if t not in self._lids])

        if binarise:
            _, statelabels = nex.characters.get_charstatelabels()
            new = CharacterMatrix.binarised(nex.characters.get_matrix(), statelabels=statelabels)
            nex.replace_block(nex.characters, Characters.from_data(new))

        normalise(nex).to_file(self.cldf_spec.dir / 'data.nex')
        self.cldf.add_provenance(
            wasDerivedFrom={
                "rdf:about": "data.nex",
                "rdf:type": "prov:Entity",
                'dc:description': 'The data underlying the analysis which created the phylogeny',
                'dc:format': 'https://en.wikipedia.org/wiki/Nexus_file',
            }
        )
        log.info("added data nexus (characters=%d)" % len(charlabels))

    def add_taxa(self,
                 taxa: typing.List[typing.Dict[str, str]],
                 glottolog,
                 log):
        glangs = {lg.id: lg for lg in glottolog.languoids()}
        #
        # FIXME: add metadata from Glottolog, put in dplace-tree-specific Dataset base class.
        # FIXME: log warnings if taxa are mapped to bookkeeping languoids!
        #
        for i, row in enumerate(taxa):
            if i == 0:
                self.add_columns(
                    'LanguageTable',
                    row,
                    log,
                    exclude=['taxon', 'glottocode', 'soc_ids', 'xd_ids'])
                self.cldf.add_columns('LanguageTable', 'Glottolog_Name')
            self._lids.add(row['taxon'])
            glang = None
            if row['glottocode']:
                try:
                    glang = glangs[row['glottocode']]
                except KeyError:  # pragma: no cover
                    log.error('Invalid glottocode in taxa.csv: {}'.format(row['glottocode']))
            d = dict(
                ID=row['taxon'],
                Name=row['taxon'],
                Glottocode=row['glottocode'] or None,
                Glottolog_Name=glang.name if glang else None,
                Latitude=glang.latitude if glang else None,
                Longitude=glang.longitude if glang else None,
            )
            self.add_obj('LanguageTable', d, row)
        log.info("added taxa (taxa=%d)" % len(taxa))



from .dataset import Dataset
from .cldfwriter import CLDFWriter
from .metadata import Metadata
from .beast import BeastFile
from .nexuslib import NexusFile
from . import commands

__version__ = '1.4.0'
__all__ = ['Dataset', 'Metadata', 'BeastFile', 'NexusFile', 'CLDFWriter']

assert commands



import copy
import typing
import zipfile
import functools
import dataclasses

import newick
from commonnexus import Nexus
from commonnexus.blocks import Trees

from .metadata import RESCALE_TO_YEARS

__all__ = ['NexusFile', 'Tree', 'rescale_to_years']


def rescale_to_years(nex: Nexus, orig_scaling, log=None) -> Nexus:
    """
    Rescales trees in a nexus file to years (if possible).

    :param nex:
    :param orig_scaling:
    :param log:
    :return: The mutated `Nexus` object.
    """
    def _rescaler(factor, n):
        n._length_formatter = lambda lg: '{:.0f}'.format(lg) if lg else None
        if n._length:
            n.length = n.length * factor

    if orig_scaling in RESCALE_TO_YEARS:
        trees = []
        for tree in nex.TREES.trees:
            nwk = tree.newick
            nwk.visit(functools.partial(_rescaler, RESCALE_TO_YEARS[orig_scaling]))
            trees.append((tree.name, nwk, tree.rooted))
        nex.replace_block(
            nex.TREES,
            Trees.from_data(*trees, **nex.TREES.TRANSLATE.mappings if nex.TREES.TRANSLATE else {}))
        return nex
    raise ValueError('Cannot rescale {} to years.'.format(orig_scaling))


@dataclasses.dataclass
class Tree:
    name: str
    newick: typing.Union[str, newick.Node]
    rooted: typing.Optional[bool] = None

    def __str__(self):
        return self.newick if isinstance(self.newick, str) else '{};'.format(self.newick.newick)


class NexusFile:
    def __init__(self, path, zipped=False):
        self.path = path
        self._trees = []
        self.scaling = None
        self.zipped = zipped

    def append(self,
               tree: typing.Union[Tree, str, newick.Node],
               tid: str,
               lids: typing.List[str],
               scaling,
               log,
               rooted: typing.Optional[bool] = None):
        if isinstance(tree, Tree):
            tid = tid or tree.name
            rooted = rooted or tree.rooted
            tree = tree.newick
        if isinstance(tree, str):
            tree = newick.loads(tree)[0]
        with_lids = bool(lids)
        if with_lids:
            lids = copy.copy(lids)
        for node in tree.walk():
            if node.name == 'root':
                continue
            if node.is_leaf:
                assert node.name
            if node.name:
                try:
                    if with_lids:
                        lids.remove(node.name)
                except KeyError:
                    if node.is_leaf:
                        log.error('{} references undefined leaf {}'.format(tree.name, node.name))
                    else:  # pragma: no cover
                        log.warning(
                            '{} references undefined inner node {}'.format(tree.name, node.name))

        if with_lids and lids:
            log.warning('extra taxa specified in LanguageTable: {}'.format(lids))

        if self.scaling:
            if scaling != self.scaling:
                raise ValueError('All trees in a NexusFile must have the same scaling!')
        else:
            self.scaling = scaling
        self._trees.append((tid, tree, rooted))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._trees:
            nex = Nexus.from_blocks(Trees.from_data(*self._trees))
            nex.to_file(self.path)
            if self.zipped:
                with zipfile.ZipFile(
                    self.path.parent / (self.path.name + '.zip'),
                    'w',
                    compression=zipfile.ZIP_DEFLATED
                ) as zf:
                    zf.write(self.path, self.path.name)
                self.path.unlink()



import collections
import urllib.parse

import attr
import cldfbench

__all__ = ['SCALING', 'RESCALE_TO_YEARS', 'Metadata']

SCALING = [
    'none',  # no branch lengths
    'change',  # parsimony steps
    'substitutions',  # change
    'years',  # years
    'centuries',  # centuries
    'millennia',  # millennia
]
ANALYSES = [
    'bayesian',
    'parsimony',
    'likelihood',
    'network',
    'other',
    'none',  # override.
]
RESCALE_TO_YEARS = {
    'centuries': 100,
    'millennia': 1000,
}


@attr.s
class Metadata(cldfbench.Metadata):
    name = attr.ib(default=None, metadata=dict(required=True))
    author = attr.ib(default=None, metadata=dict(required=True))
    year = attr.ib(default=None)
    scaling = attr.ib(
        default='none',
        validator=attr.validators.in_(SCALING),
        metadata=dict(required=True))
    analysis = attr.ib(
        default='none',
        validator=attr.validators.in_(ANALYSES),
        metadata=dict(required=True))
    family = attr.ib(default=None, metadata=dict(required=True))
    cldf = attr.ib(
        default=None,
        converter=lambda s: 'https://{}'.format(s) if s and s.startswith('github.com') else s)
    data = attr.ib(default=None)
    missing = attr.ib(default=attr.Factory(dict))
    zenodo_concept_doi = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.url:
            u = urllib.parse.urlparse(self.url)
            if u.netloc == 'dx.doi.org':
                self.url = urllib.parse.urlunsplit(('https', 'doi.org', u.path, '', ''))

        ref = self.author or ''
        if self.year:
            ref += ' {}'.format(self.year)
        if self.name:
            ref += " '{}'".format(self.name.strip())
        if ref:
            ref = 'derived from {}'.format(ref)
        else:
            ref = self.id
        self.title = "Phlorest phylogeny {}".format(ref)

    def common_props(self):
        res = cldfbench.Metadata.common_props(self)
        res['dc:subject'] = collections.OrderedDict()
        for k in ['family', 'analysis', 'scaling']:
            v = getattr(self, k)
            if v:
                res['dc:subject'][k] = v
        data = self.cldf or self.data
        if data:
            res['prov:wasDerivedFrom'] = [{
                "rdf:about": data,
                "rdf:type": "prov:Entity",
                "dc:description": "Dataset underlying the analysis"
            }]
        return res

    @property
    def text_description(self):
        res = 'A [Phlorest phylogeny](https://github.com/phlorest)'
        if self.family:
            if self.family == 'Multiple':  # pragma: no cover
                res += ' of multiple language families'
            else:
                res += ' of the {} language family'.format(self.family)
        if self.analysis and self.analysis != 'none':
            res += ' computed from a {} analysis'.format(self.analysis)
        if self.scaling and self.scaling != 'none':
            res += ' scaled by {}'.format(self.scaling)
        return res + '.'



import gzip
import shlex
import shutil
import random
import typing
import pathlib
import subprocess

import cldfbench
from cldfbench.datadir import DataDir
from pyglottolog.languoids import Glottocode
from clldutils.path import TemporaryDirectory, ensure_cmd
from pycldf.trees import TreeTable
from cldfviz.tree import render
from commonnexus import Nexus
from commonnexus.tools.normalise import normalise as nexus_norm

from .nexuslib import Tree
from .metadata import Metadata
from .cldfwriter import CLDFWriter


class PhlorestDir(DataDir):
    """
    Enhanced `DataDir`, adding methods to access phylogenetic data.
    """
    def read_nexus(self,
                   path: typing.Optional[typing.Union[str, pathlib.Path]] = None,
                   text: typing.Optional[str] = None,
                   encoding: str = 'utf-8-sig',
                   normalise: bool = False,
                   preprocessor: typing.Callable[[str], str] = lambda s: s) -> Nexus:
        """
        :param path: path to nexus file (or `None`).
        :param text: text content of a nexus file.
        :return: Initialized `Nexus` object.
        """
        assert (path or text) and not (path and text), 'Must pass either path or text'
        if path:
            path = self._path(path)
            if path.suffix == '.gz':
                with gzip.open(path, 'rt', encoding='utf8') as fp:
                    text = fp.read()
        res = Nexus(preprocessor(text or self.read(path, encoding=encoding)))
        return nexus_norm(res) if normalise else res

    def read_trees(self,
                   path: typing.Union[str, pathlib.Path] = None,
                   text: str = None,
                   detranslate: bool = False,
                   burnin: int = 0,
                   sample: int = 0,
                   strip_annotation: bool = False,
                   seed=12345,
                   preprocessor: typing.Callable[[str], str] = lambda s: s) -> typing.List[Tree]:
        """
        Reads trees from `path` and transforms them as required.

        Processing order:
            burnin -> sample -> strip_annotation -> remove_rate -> detranslate

        :param path: path to nexus file.
        :param text: nexus content in text.
        :param detranslate: return trees with translate blocks removed (default=False).
        :param burnin: number of trees to remove as burn-in (default=none).
        :param sample: number of trees to sample (default=all).
        :param remove_rate: remove extra rate information.
        :param strip_annotation: remove comments and annotations in trees (default=False).
        :param preprocessor: function to preprocess nexus text.
        :return:
        """
        nex = self.read_nexus(path=path, text=text, preprocessor=preprocessor)
        trees = nex.TREES.trees
        # remove burn-in first
        if burnin:
            trees = trees[burnin:]
        # ..then sample if needed
        if sample and len(trees) > sample:
            random.seed(seed)
            trees = random.sample(trees, sample)

        trees = [Tree(tree.name, tree.newick, tree.rooted) for tree in trees]
        # ...then detranslate.
        if detranslate:
            # We must use a reference to the same block in order to make the translation-mapping
            # caching work.
            cmd = nex.TREES.translate
            for tree in trees:
                tree.newick = cmd(tree.newick)

        # remove comments if asked
        if strip_annotation:
            for tree in trees:
                tree.newick.strip_comments()

        return trees

    def read_tree(self,
                  path: typing.Union[str, pathlib.Path] = None,
                  text: str = None,
                  detranslate: bool = False,
                  burnin: int = 0,
                  sample: int = 0,
                  remove_rate: bool = False,
                  strip_annotation: bool = False,
                  seed=12345,
                  preprocessor=lambda s: s) -> Tree:
        return self.read_trees(
            path=path,
            text=text,
            detranslate=detranslate,
            burnin=burnin,
            sample=sample,
            strip_annotation=strip_annotation,
            seed=seed,
            preprocessor=preprocessor)[0]


class Dataset(cldfbench.Dataset):
    """
    An augmented `cldfbench.Dataset`

    - swapping in `PhlorestDir` as `DataDir` implementation for `raw`
    - swapping in a custom CLDFWriter implementation `phorest.cldfwriter.CLDFWriter`
    - adding methods to be called in implementations of `cmd_makecldf` for simpler manipulation of \
      phylogenetic data,
    - enhancing README.md by adding an SVG plot of the summary tree.
    """
    metadata_cls = Metadata
    datadir_cls = PhlorestDir

    def __init__(self):
        cldfbench.Dataset.__init__(self)
        self._lids = set()

    def cldf_specs(self):  # A dataset must declare all CLDF sets it creates.
        return cldfbench.CLDFSpec(dir=self.cldf_dir, writer_cls=CLDFWriter)

    def cmd_download(self, args):  # pragma: no cover
        pass

    def _cmd_makecldf(self, args):
        if self.metadata.family and Glottocode.pattern.match(self.metadata.family):
            glang = args.glottolog.api.languoid(self.metadata.family)
            self.metadata.family = '{} [{}]'.format(glang.name, glang.id)
        cldfbench.Dataset._cmd_makecldf(self, args)

        cldf = self.cldf_reader()
        for tree in TreeTable(cldf):
            if tree.tree_type == 'summary':
                legend = "Summary tree"
                if cldf.properties.get('dc:subject', {}).get('analysis'):
                    title = cldf.properties['dc:subject']['analysis'].title()
                    legend += ' of a {} analysis'.format(title)
                if cldf.properties.get('dc:subject', {}).get('family'):
                    family = cldf.properties['dc:subject']['family']
                    legend += ' of the {} family'.format(family)
                if tree.tree_branch_length_unit:
                    legend += ' with branches in {}'.format(tree.tree_branch_length_unit)

                return render(
                    tree,
                    self.dir / 'summary_tree.svg',
                    glottolog_mapping={
                        r['ID']: (r['Glottocode'], r.get('Glottolog_Name'))
                        for r in cldf['LanguageTable'] if r['Glottocode']},
                    legend=legend,
                    width=1000,
                    with_glottolog_links=True
                )

    def init(self, args):
        """
        Create rows in LanguageTable according to `etc/taxa.csv` and add sources from
        `raw/sources.bib`.
        """
        args.writer.add_taxa(self.taxa, args.glottolog.api, args.log)
        if self.raw_dir.joinpath('sources.bib').exists():
            args.writer.cldf.sources.add(
                self.raw_dir.joinpath('sources.bib').read_text(encoding='utf8'))

    def _cmd_readme(self, args):
        cldfbench.Dataset._cmd_readme(self, args)
        text = self.dir.joinpath('README.md').read_text(encoding='utf8')
        text = text.replace('Available online', 'Source available online')
        pre, header, post = text.partition('## Description')
        text = pre + header + '\n\n' + self.metadata.text_description + post

        lines = []
        for line in text.split('\n'):
            lines.append(line)
            if line.startswith('[![CLDF validation]') and self.metadata.zenodo_concept_doi:
                lines.append(
                    '[![DOI](https://zenodo.org/badge/DOI/{0}.svg)](https://doi.org/{0})'.format(
                        self.metadata.zenodo_concept_doi))  # pragma: no cover
        text = '\n'.join(lines)

        if self.dir.joinpath('summary_tree.svg').exists():
            text += '\n\n## Summary Tree\n\n!' \
                    '[summary](https://raw.githubusercontent.com/phlorest/{}/' \
                    'main/summary_tree.svg)'.format(self.id)
        self.dir.joinpath('README.md').write_text(text, encoding='utf8')
        print('gh repo edit --description "{}" --add-topic "phylogeny"'.format(self.metadata.title))
        if self.metadata.family:
            print('gh repo edit --add-topic "language-family-{}"'.format(
                self.metadata.family.lower().replace(' ', '')))
        if self.metadata.url:  # pragma: no cover
            print('gh repo edit --homepage "{}"'.format(self.metadata.url))

    def _read_from_etc(self, name):
        if (self.etc_dir / name).exists():
            return list(self.etc_dir.read_csv(name, dicts=True))
        return []

    @property
    def taxa(self) -> typing.List[dict]:
        return self._read_from_etc('taxa.csv')

    @property
    def characters(self) -> typing.List[dict]:
        return self._read_from_etc('characters.csv')

    @staticmethod
    def run_treeannotator(cmd, input: typing.Union[str, pathlib.Path]) -> Nexus:
        with TemporaryDirectory() as d:
            in_ = d / 'in.nex'
            if isinstance(input, str):
                in_.write_text(input, encoding='utf8')
            else:
                shutil.copy(input, in_)
            out = d / 'out.nex'
            subprocess.check_call(
                [ensure_cmd('treeannotator')] + shlex.split(cmd) + [str(in_), str(out)],
                stderr=subprocess.DEVNULL,
            )
            return Nexus(out.read_text(encoding='utf8'))

    @staticmethod
    def run_rscript(script, output_fname):
        with TemporaryDirectory() as d:
            d.joinpath('script.r').write_text(script, encoding='utf8')
            subprocess.check_call([ensure_cmd('Rscript'), str(d / 'script.r')], cwd=d)
            return d.joinpath(output_fname).read_text(encoding='utf8')



import pathlib

from cldfbench.scaffold import Template

import phlorest


class PhlorestTemplate(Template):
    package = 'phlorest'

    dirs = Template.dirs + [pathlib.Path(phlorest.__file__).parent / 'phlorest_template']
    metadata = phlorest.Metadata



import shutil
import logging
import argparse

from phlorest.commands import check, contrib
from phlorest.__main__ import main
from phlorest import Dataset


def test_check(dataset, caplog):
    check.run(argparse.Namespace(log=logging.getLogger(__name__)), dataset)
    assert len(caplog.records) >= 4


def test_contrib(dataset, capsys):
    contrib.run(argparse.Namespace(log=logging.getLogger(__name__)), dataset)
    out, _ = capsys.readouterr()
    assert 'GitHub user' in out


def test_check_characters(tmp_repos):
    tmp_repos.joinpath('etc', 'characters.csv').write_text('a,b\n,', encoding='utf8')
    shutil.rmtree(tmp_repos / 'cldf')

    class DS(Dataset):
        dir = tmp_repos
        id = 'phy'

    check.run(argparse.Namespace(log=logging.getLogger(__name__)), DS())


def test_main(dataset):
    main(parsed_args=argparse.Namespace(dataset=dataset))


def test_valid(cldf_dataset, cldf_logger):
    assert cldf_dataset.validate(log=cldf_logger)


def test_phlorest_check(cldf_dataset, cldf_logger):
    from phlorest.check import run_checks
    assert run_checks(cldf_dataset, cldf_logger)



import shutil
import argparse

import pytest

from phlorest.dataset import PhlorestDir


@pytest.fixture
def cldfwriter(dataset):
    with dataset.cldf_writer(argparse.Namespace()) as w:
        return w


def test_PhlorestDir(repos):
    d = PhlorestDir(repos / 'raw')
    nex = d.read_nexus('nexus.trees', preprocessor=lambda s: s.replace('Cojubim', 'abcdefg'))
    assert 'abcdefg' in str(nex)

    nex = d.read_nexus('nexus.trees.gz')
    assert nex.TREES.TREE


def test_PhlorestDir_with_rates(repos):
    d = PhlorestDir(repos / 'raw')
    nex = d.read_nexus('trees_with_rate.trees')
    for node in nex.TREES.TREE.newick.walk():
        if node.name == '1':
            assert ('height' in node.properties) and ('rate' in node.properties)


@pytest.mark.noci
def test_Dataset(dataset, cldfwriter, mocker, glottolog, tmp_path):
    assert dataset.metadata.title == "Phlorest phylogeny derived from the author 2021 'The name'"
    args = argparse.Namespace(
        writer=cldfwriter, glottolog=mocker.Mock(api=glottolog), log=mocker.Mock())
    dataset._cmd_makecldf(args)
    dataset._cmd_readme(args)


def test_Dataset_run_treeannotator(dataset, mocker, repos):
    def annotate(args, **kw):
        shutil.copy(repos / 'raw' / 'nexus.trees', args[-1])

    mocker.patch('phlorest.dataset.ensure_cmd', mocker.Mock(return_value='cmd'))
    mocker.patch('phlorest.dataset.shutil', mocker.Mock())
    mocker.patch('phlorest.dataset.subprocess', mocker.Mock(check_call=annotate))
    _ = dataset.run_treeannotator('cmd', 'in')
    res = dataset.run_treeannotator('cmd', dataset.raw_dir / 'nexus.trees')
    assert res.TREES


def test_Dataset_run_rscript(dataset, mocker):
    def rscript(args, **kw):
        kw['cwd'].joinpath('res.txt').write_text('hello', encoding='utf8')

    mocker.patch('phlorest.dataset.subprocess', mocker.Mock(check_call=rscript))
    res = dataset.run_rscript('', 'res.txt')
    assert res == 'hello'


def test_PhlorestDir_read_trees(dataset):
    tfile = dataset.raw_dir / 'posterior.trees'
    # no args
    assert len(dataset.raw_dir.read_trees(tfile)) == 3
    
    # sample
    assert len(dataset.raw_dir.read_trees(tfile, sample=2)) == 2
    
    # burnin
    trees = dataset.raw_dir.read_trees(tfile, burnin=2)
    assert len(trees) == 1
    assert 'TREE3' == trees[0].name

    trees = dataset.raw_dir.read_trees(tfile, burnin=1)
    assert len(trees) == 2
    assert 'TREE2' == trees[0].name
    assert 'TREE3' == trees[1].name
    
    # strip_annotation
    trees = dataset.raw_dir.read_trees(tfile, burnin=2, strip_annotation=True)
    assert len(trees) == 1
    assert 'TREE3' == trees[0].name
    assert '[' not in trees[0].newick.newick

    # preprocessor
    trees = dataset.raw_dir.read_trees(
        tfile,
        burnin=2,
        preprocessor=lambda t: t.replace("tree TREE", "tree TESTTREE")
    )
    assert len(trees) == 1
    assert 'TESTTREE3' == trees[0].name
    
    # detranslate
    trees = dataset.raw_dir.read_trees(tfile, burnin=2, detranslate=True)
    assert len(trees) == 1
    assert 'Cojubim' in trees[0].newick.newick
    
    # combined
    trees = dataset.raw_dir.read_trees(
        tfile,
        burnin=1, 
        sample=1,
        preprocessor=lambda t: t.replace("tree TREE", "tree TESTTREE"),
        strip_annotation=True,
        detranslate=True
    )
    assert len(trees) == 1
    assert 'TREE1' != trees[0].name, 'Tree1 should never be sampled due to burn-in setting'
    assert '[' not in trees[0].newick.newick
    assert 'Cojubim' in trees[0].newick.newick



"""
Generates CONTRIBUTORS.md from raw/sources.bib
"""
from cldfbench.cli_util import add_dataset_spec, get_dataset


def register(parser):  # pragma: no cover
    add_dataset_spec(parser)


def person_to_str(p):  # inverse of pybtex's str()
    von_last = ' '.join(p.prelast_names + p.last_names)
    jr = ' '.join(p.lineage_names)
    first = ' '.join(p.first_names + p.middle_names)
    return ' '.join(part for part in (first, jr, von_last) if part)


def run(args, d=None):
    if d is None:  # pragma: no cover
        try:
            d = get_dataset(args)
        except Exception as e:
            args.log.error("Unable to load %s - %s" % (args.dataset, e))
            raise

    # assume first source is this dataset
    print("%-30s| GitHub user | Description | Role" % 'Name')
    print("%-30s| ----------- | ----------- | ----" % ('-' * 30))
    for author in d.raw_dir.read_bib()[0].entry.persons['author']:
        print("%-30s|             | author      | Author" % person_to_str(author))



import typing

from pycldf import Dataset as CLDFDataset

import attr
from cldfbench.dataset import dataset_from_module

from phlorest import Dataset, Metadata

# values in metadata.json that should be present and should not be empty
METAKEYS = [a.name for a in attr.fields(Metadata) if a.metadata.get('required')]


def run_checks(d: typing.Union[CLDFDataset, Dataset], log) -> bool:
    """
    Run a couple of Phlorest-specific checks on a dataset.

    To invoke from a dataset's `test.py` module, include

    .. code-block:: python

        def test_phlorest_check(cldf_dataset, cldf_logger):
            from phlorest.check import run_checks
            assert run_checks(cldf_dataset, cldf_logger)

    :return: `True` if all checks passed, `False` otherwise.
    """
    if isinstance(d, CLDFDataset):
        d = dataset_from_module(
            d.directory.joinpath('..', 'cldfbench_{}.py'.format(d.properties['rdf:ID'])))

    success = True

    def check(condition, msg):
        if condition:
            log.warning('{0.id}: {1}'.format(d, msg))
            return False
        return True

    # check metadata
    for mdkey in METAKEYS:
        success &= check(
            not getattr(d.metadata, mdkey, ''),
            "metadata missing value for `%s`" % mdkey)

    success &= check(
        not (d.raw_dir / 'sources.bib').exists(),
        "raw/sources.bib file missing")

    success &= check(
        not (d.dir / 'CONTRIBUTORS.md').exists(),
        "CONTRIBUTORS.md file missing")

    success &= check(
        not ((d.cldf_dir / 'summary.trees').exists() or d.metadata.missing.get('summary')),
        "missing summary tree not declared")

    success &= check(
        not ((d.cldf_dir / 'posterior.trees.zip').exists() or d.metadata.missing.get('posterior')),
        "missing posterior tree not declared")

    success &= check(
        not ((d.cldf_dir / 'data.nex').exists() or d.metadata.missing.get('nexus')),
        "missing nexus data not declared")

    success &= check(
        not (d.characters or d.metadata.missing.get('characters')),
        "missing characters.csv not declared")

    success &= check(
        any('concepticonReference' in char for char in d.characters),
        "characters.csv uses `concepticonReference` rather than `Concepticon_ID`")

    # check that characters are coded if possible
    if d.characters and not d.metadata.missing.get('concepticon'):
        success &= check(
            all(char.get('Concepticon_ID', "") == "" for char in d.characters),
            "characters.csv file missing concepticon coding")

    success &= check((d.dir / 'Makefile').exists(), "has a legacy Makefile")
    return success



import cldfbench

from phlorest.cldfwriter import CLDFWriter
from phlorest.metadata import Metadata
from phlorest.beast import BeastFile


def test_CLDFWriter(repos, tmp_path, mocker, nexus_tree, dataset, glottolog):
    log = mocker.Mock()
    with CLDFWriter(cldf_spec=cldfbench.CLDFSpec(dir=tmp_path)) as writer:
        writer.add_taxa(dataset.taxa, glottolog, mocker.Mock())
        writer.cldf.add_table('test.csv', 'a')
        writer.add_columns('test.csv', {'a': 'b', 'c': 3}, log=log)
        assert log.error.called
        assert writer.cldf['test.csv', 'a']
        writer.add_summary(
            nexus_tree,
            Metadata(name='n', author='a', year=2021),
            mocker.Mock())
        writer.add_posterior(
            [nexus_tree],
            Metadata(name='n', author='a', year=2021),
            mocker.Mock())
        writer.add_data(BeastFile(repos / 'raw' / 'beast.xml'), [], mocker.Mock())


def test_CLDFWriter_nexus_data(repos, tmp_path, mocker, nexus_tree, dataset, glottolog):
    with CLDFWriter(cldf_spec=cldfbench.CLDFSpec(dir=tmp_path)) as writer:
        writer.add_taxa(dataset.taxa, glottolog, mocker.Mock())
        writer.add_data(repos / 'raw' / 'data.nex', [{'Site': '0', 'Gloss': 'abc'}], mocker.Mock())
        assert writer.cldf['ParameterTable', 'Gloss']


def test_CLDFWriter_nexus_data_2(repos, tmp_path, mocker, nexus_tree, dataset, glottolog):
    with CLDFWriter(cldf_spec=cldfbench.CLDFSpec(dir=tmp_path)) as writer:
        writer.add_taxa(dataset.taxa, glottolog, mocker.Mock())
        writer.add_data(
            (repos / 'raw' / 'data.nex').read_text(encoding='utf8'),
            [{'Site': '0', 'Gloss': 'abc'}], mocker.Mock())
        assert writer.cldf['ParameterTable', 'Gloss']

    with CLDFWriter(cldf_spec=cldfbench.CLDFSpec(dir=tmp_path)) as writer:
        writer.add_taxa(dataset.taxa, glottolog, mocker.Mock())
        writer.add_data(
            """\
#NEXUS

BEGIN DATA;
	dimensions ntax=3 nchar=1;
	format datatype=STANDARD gap=- missing=? symbols="ABC";
	MATRIX
	Jeju A
	SouthJeolla B
	NorthJeolla C
	;
END;""",
            [{'Site': '0', 'Gloss': 'abc'}],
            mocker.Mock(),
            binarise=True)
        assert writer.cldf['ParameterTable', 'Gloss']


def test_CLDFWriter_nexus_data_3(repos, tmp_path, mocker, nexus_tree, dataset, glottolog):
    from commonnexus import Nexus

    with CLDFWriter(cldf_spec=cldfbench.CLDFSpec(dir=tmp_path)) as writer:
        writer.add_taxa(dataset.taxa, glottolog, mocker.Mock())
        writer.add_data(
            Nexus((repos / 'raw' / 'data.nex').read_text(encoding='utf8')),
            [{'Site': '0', 'Gloss': 'abc'}], mocker.Mock())
        assert writer.cldf['ParameterTable', 'Gloss']



import typing
import pathlib
import collections
import xml.etree.cElementTree as ElementTree

from commonnexus import Nexus
from commonnexus.blocks import Characters

__all__ = ['BeastFile']


class BeastFile:
    def __init__(self, path: typing.Union[str, pathlib.Path], text: typing.Optional[str] = None):
        self.path = path
        self.text = text
        if self.text:
            self.xml = ElementTree.fromstring(self.text)
        else:
            self.xml = ElementTree.parse(str(self.path))

    def nexus(self, valid_states: str = '01?') -> Nexus:
        """
        Read the character data used in a BEAST file and format it as NEXUS file.

        :param valid_states: String listing one-character state labels.
        :return: A Nexus instance with a CHARACTERS block encoding the data from the BEAST file as \
        MATRIX.
        """
        try:  # Read the character labels.
            chars = dict(self.iter_characters())
        except (ValueError, KeyError):  # pragma: no cover
            chars = {}  # No character labels.

        matrix = collections.OrderedDict()

        def add_row(taxon, seq):
            matrix[taxon] = collections.OrderedDict()
            for i, state in enumerate([s for s in seq if s != ' '], start=1):
                assert state in valid_states, 'Invalid State %s' % state
                matrix[taxon][chars.get(i, str(i))] = None if state == '?' else state

        for seq in self.xml.findall('./data/sequence'):
            add_row(seq.get('taxon'), seq.get('value'))

        if not matrix:
            for seq in self.xml.findall('.//sequence[taxon]'):
                data = (seq.text.strip() if seq.text else None) or seq.find('taxon').tail.strip()
                assert data, ElementTree.tostring(seq).decode('utf8').replace('\n', '')
                add_row(seq.find('taxon').attrib['idref'], data)

        return Nexus.from_blocks(Characters.from_data(matrix))

    def iter_characters(self):
        def get_partition(p):
            x, y = [int(_) for _ in p.get('filter').split("-")]
            return (p.get('id'), x, y)

        def printchar(p, x, y, ascertained=False):
            n = 1
            for i in range(x, y + 1):
                label = "%s-%s" % (p, 'ascertained' if n == 1 and ascertained else str(n))
                yield i, label
                n += 1

        def get_by_id(data_id):
            if data_id.startswith("@"):
                data_id = data_id.lstrip("@")
            res = self.xml.find(".//alignment[@id='%s']" % data_id)
            if res is None:  # pragma: no cover
                raise ValueError(data_id)
            return res

        for treelh in self.xml.findall(".//distribution[@spec='TreeLikelihood']"):
            if treelh.get('data'):
                data = get_by_id(treelh.get('data'))
                ascertained = data.get('ascertained') == 'true'
                yield from printchar(*get_partition(data.find('./data')), ascertained=ascertained)
            else:
                data = treelh.find('./data')
                ascertained = data.get('ascertained') == 'true'
                dd = get_by_id(data.get('data')) if data.get('data') else treelh.find('./data/data')
                yield from printchar(*get_partition(dd), ascertained=ascertained)



import pytest

from commonnexus import Nexus

from phlorest.nexuslib import NexusFile, rescale_to_years, Tree


def test_rescale_to_years():
    nex = Nexus("""#NEXUS
begin trees;
tree t1 = (A:258,B:123)C:254;
end;""")
    res = rescale_to_years(nex, 'millennia')
    assert '258000' in res.TREES.TREE.newick_string

    with pytest.raises(ValueError):
        _ = rescale_to_years(nex, 'millenia')


def test_NexusFile(tmp_path, mocker):
    with NexusFile(tmp_path / 'test.nex') as nex:
        nex.append(Tree('n', '(A:1,B:2)root:3;', None), 'a', {'A', 'B'}, 'years', mocker.Mock())
        with pytest.raises(ValueError):
            nex.append(Tree('n', '(A:1,B:2)root:3;', None), 'c', {'A', 'B'}, 'change', mocker.Mock())
    res = Nexus.from_file(tmp_path / 'test.nex')
    assert len(res.TREES.trees) == 1


def test_Tree():
    t = Tree('n', '(A:1,B:2)root:3;', None)
    assert str(t) == '(A:1,B:2)root:3;'



import logging

from phlorest.check import run_checks


def test_run_checks(dataset, caplog):
    assert run_checks(dataset.cldf_reader(), logging.getLogger(__name__)) is False
    assert len(caplog.records) == 5



from phlorest import BeastFile


def test_BeastFile_1(dataset):
    bf = BeastFile(None, text=dataset.raw_dir.read('beast2.xml.gz'))
    nex = bf.nexus()
    matrix = nex.characters.get_matrix()
    assert list(matrix['A'])[135] == 'again-62'  # check one
    assert sorted(nex.taxa) == ['A', 'B', 'C']
    assert len(matrix['A']) == 235


def test_BeastFile_2(dataset):
    bf = BeastFile(None, text=dataset.raw_dir.read('beast.xml'))
    nex = bf.nexus()
    matrix = nex.characters.get_matrix()
    assert nex.taxa == ['Jeju', 'SouthJeolla', 'NorthJeolla']
    assert len(matrix['Jeju']) == 384
    assert list(matrix['Jeju'])[1] == '2'  # no character labels



"""

"""
from cldfbench.cli_util import add_dataset_spec, get_dataset


def register(parser):
    add_dataset_spec(parser)
    parser.add_argument('tag')


def run(args):  # pragma: no cover
    tag = args.tag
    if not tag.startswith('v'):
        tag = 'v' + tag
    ds = get_dataset(args)
    cldf = ds.cldf_reader()
    ds.dir.joinpath('relnotes.txt').write_text(
        """\
Cite the source as

> {}

and the Phlorest phylogeny as

DOI""".format(cldf.properties['dc:bibliographicCitation']),
        encoding='utf8')
    print('gh release create {} --title "{}" --notes-file relnotes.txt'.format(
        tag, cldf.properties['dc:title']))
    print('')
    print("Now you should grab the Zenodo version DOI from\n"
          "https://zenodo.org/account/settings/github/repository/phlorest/{0}\n"
          "and add it to\n"
          "https://github.com/phlorest/{0}/releases/edit/{1}\n and the concept DOI to "
          "metdata.json".format(cldf.properties['rdf:ID'], tag))
